# Autogenerated by Astropy-affiliated package astropy_helpers's setup.py on 2016-01-08 16:04:15.594755
from __future__ import unicode_literals
import datetime

version = "1.1.1"
githash = "fedd7c52ba80e2246f83b8d1752baa3a153a33e2"


major = 1
minor = 1
bugfix = 1

release = True
timestamp = datetime.datetime(2016, 1, 8, 16, 4, 15, 594755)
debug = False

try:
    from ._compiler import compiler
except ImportError:
    compiler = "unknown"

try:
    from .cython_version import cython_version
except ImportError:
    cython_version = "unknown"
