# Autogenerated by Astropy's setup.py on 2016-01-08 16:04:16.052991
from __future__ import unicode_literals
import datetime

version = "1.1.1"
githash = "6529f253aa10a19f6da58868299392034173c1c0"


major = 1
minor = 1
bugfix = 1

release = True
timestamp = datetime.datetime(2016, 1, 8, 16, 4, 16, 52991)
debug = False

try:
    from ._compiler import compiler
except ImportError:
    compiler = "unknown"

try:
    from .cython_version import cython_version
except ImportError:
    cython_version = "unknown"
