# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import (absolute_import, division, print_function, unicode_literals)

_tabversion   = '3.5'
_lextokens    = set(['DEGREE', 'HOUR', 'SIMPLE_UNIT', 'SIGN', 'SECOND', 'COLON', 'UINT', 'MINUTE', 'UFLOAT'])
_lexreflags   = 0
_lexliterals  = ''
_lexstateinfo = {'INITIAL': 'inclusive'}
_lexstatere   = {'INITIAL': [('(?P<t_UFLOAT>((\\d+\\.\\d*)|(\\.\\d+))([eE][+-\u2212]?\\d+)?)|(?P<t_UINT>\\d+)|(?P<t_SIGN>[+\u2212-])|(?P<t_SIMPLE_UNIT>(?:Gdeg)|(?:hectoradian)|(?:petaradian)|(?:decaarcminute)|(?:hrad)|(?:zeptoradian)|(?:arcsec)|(?:aarcmin)|(?:Marcmin)|(?:decadegree)|(?:kiloarcsecond)|(?:exaarcsecond)|(?:Parcsec)|(?:hectodegree)|(?:yoctoarcminute)|(?:prad)|(?:ydeg)|(?:Parcmin)|(?:yottaradian)|(?:marcmin)|(?:aarcsec)|(?:zettadegree)|(?:milliradian)|(?:parcsec)|(?:zdeg)|(?:attodegree)|(?:arad)|(?:millidegree)|(?:uarcmin)|(?:parcmin)|(?:nanodegree)|(?:petadegree)|(?:milliarcminute)|(?:kilodegree)|(?:zeptodegree)|(?:exadegree)|(?:radian)|(?:microdegree)|(?:dekaarcsecond)|(?:nrad)|(?:kdeg)|(?:Tarcsec)|(?:attoarcsecond)|(?:Trad)|(?:krad)|(?:Zarcmin)|(?:Earcmin)|(?:exaarcminute)|(?:farcmin)|(?:gigadegree)|(?:Prad)|(?:dekaarcminute)|(?:Earcsec)|(?:udeg)|(?:Yarcsec)|(?:dadeg)|(?:Garcmin)|(?:daarcmin)|(?:uarcsec)|(?:kiloradian)|(?:nanoarcminute)|(?:kiloarcminute)|(?:Zdeg)|(?:megaradian)|(?:yarcmin)|(?:Zarcsec)|(?:carcsec)|(?:femtoarcsecond)|(?:Marcsec)|(?:mdeg)|(?:centidegree)|(?:zettaarcsecond)|(?:yoctoarcsecond)|(?:arcsecond)|(?:decaarcsecond)|(?:zeptoarcminute)|(?:attoradian)|(?:Edeg)|(?:Grad)|(?:Garcsec)|(?:Erad)|(?:deciarcminute)|(?:yoctodegree)|(?:picoarcminute)|(?:ndeg)|(?:teraradian)|(?:hdeg)|(?:narcsec)|(?:adeg)|(?:teraarcsecond)|(?:zettaradian)|(?:darad)|(?:microradian)|(?:gigaarcsecond)|(?:milliarcsecond)|(?:decaradian)|(?:exaradian)|(?:centiarcminute)|(?:cy)|(?:ddeg)|(?:femtoradian)|(?:mrad)|(?:femtoarcminute)|(?:yottaarcsecond)|(?:centiarcsecond)|(?:deciradian)|(?:Tdeg)|(?:zettaarcminute)|(?:Mdeg)|(?:picodegree)|(?:yoctoradian)|(?:decidegree)|(?:crad)|(?:picoradian)|(?:gigaradian)|(?:zrad)|(?:dekaradian)|(?:femtodegree)|(?:pdeg)|(?:narcmin)|(?:daarcsec)|(?:microarcminute)|(?:hectoarcminute)|(?:urad)|(?:nanoarcsecond)|(?:microarcsecond)|(?:hectoarcsecond)|(?:petaarcsecond)|(?:deciarcsecond)|(?:arcminute)|(?:nanoradian)|(?:Mrad)|(?:carcmin)|(?:dekadegree)|(?:yottaarcminute)|(?:rad)|(?:uas)|(?:harcmin)|(?:fdeg)|(?:megaarcsecond)|(?:Ydeg)|(?:yrad)|(?:Zrad)|(?:Yarcmin)|(?:zeptoarcsecond)|(?:zarcmin)|(?:darcmin)|(?:marcsec)|(?:karcmin)|(?:attoarcminute)|(?:farcsec)|(?:Tarcmin)|(?:teradegree)|(?:yarcsec)|(?:mas)|(?:frad)|(?:petaarcminute)|(?:teraarcminute)|(?:harcsec)|(?:yottadegree)|(?:Pdeg)|(?:gigaarcminute)|(?:drad)|(?:arcmin)|(?:centiradian)|(?:darcsec)|(?:zarcsec)|(?:cycle)|(?:cdeg)|(?:karcsec)|(?:megaarcminute)|(?:megadegree)|(?:picoarcsecond)|(?:Yrad))|(?P<t_SECOND>s(ec(ond(s)?)?)?|\u2033|\\"|\u02e2)|(?P<t_MINUTE>m(in(ute(s)?)?)?|\u2032|\\\'|\u1d50)|(?P<t_DEGREE>d(eg(ree(s)?)?)?|\xb0)|(?P<t_HOUR>hour(s)?|h(r)?|\u02b0)|(?P<t_COLON>:)', [None, ('t_UFLOAT', 'UFLOAT'), None, None, None, None, ('t_UINT', 'UINT'), ('t_SIGN', 'SIGN'), ('t_SIMPLE_UNIT', 'SIMPLE_UNIT'), (None, 'SECOND'), None, None, None, (None, 'MINUTE'), None, None, None, (None, 'DEGREE'), None, None, None, (None, 'HOUR'), None, None, (None, 'COLON')])]}
_lexstateignore = {'INITIAL': ' '}
_lexstateerrorf = {'INITIAL': 't_error'}
_lexstateeoff = {}
